from typing import Optional, Union, List

from ..math.matrix import Matrix
from ..dynamics.body import Body
from ..common.random import RandomGenerator
from .joint.joint import Joint
from .joint.distance import DistanceJoint, DistanceJointPrimitive
from .joint.point import PointJoint, PointJointPrimitive
from .joint.pulley import PulleyJoint, PulleyJointPrimitive
from .joint.revolute import RevoluteJoint, RevoluteJointPrimitive
from .joint.rotation import OrientationJoint, OrientationJointPrimitive
from .joint.rotation import RotationJointPrimitive, RotationJoint


class PhysicsWorld():
    def __init__(self):
        self._gravity: Matrix = Matrix([0.0, -1.0], 'vec')
        self._linear_vel_damping: float = 0.9
        self._ang_vel_damping: float = 0.9
        self._linear_vel_threshold: float = 0.02
        self._ang_vel_threshold: float = 0.02
        self._air_fric_coeff: float = 0.7

        self._bias: float = 0.8
        self._vel_iter: int = 1
        self._pos_iter: int = 1

        self._grav_ena: bool = True
        self._damping_ena: bool = True
        self._body_list: List[Body] = []
        self._joint_list: List[Joint] = []

    def prepare_velocity_constraint(self, dt: float) -> None:
        for joint in self._joint_list:
            if joint.active:
                joint.prepare(dt)

    def step_velocity(self, dt: float) -> None:
        g: Matrix = self._gravity if self._grav_ena else Matrix([0.0, 0.0],
                                                                'vec')
        lvd: float = 1.0
        avd: float = 1.0

        if self._damping_ena:
            lvd = 1.0 / (1.0 + dt * self._linear_vel_damping)
            avd = 1.0 / (1.0 + dt * self._ang_vel_damping)

        for body in self._body_list:
            if body.type == Body.Type.Static:
                body.vel.clear()
                body.ang_vel = 0.0

            elif body.type == Body.Type.Dynamic:
                body.forces += g * body.mass
                # NOTE: meet the operator overload seq
                body.vel += body.forces * dt * body.inv_mass
                body.ang_vel += body.inv_inertia * body.torques * dt

                body.vel *= lvd
                body.ang_vel *= avd

            elif body.type == Body.Type.Kinematic:
                # NOTE: meet the operator overload seq
                body.vel += body.forces * dt * body.inv_mass
                body.ang_vel += body.inv_inertia * body.torques * dt

                body.vel *= lvd
                body.ang_vel *= avd

            elif body.type == Body.Type.Bullet:
                pass

    def solve_velocity_constraint(self, dt: float) -> None:
        for joint in self._joint_list:
            if joint.active:
                joint.solve_velocity(dt)

    def step_position(self, dt: float) -> None:
        for body in self._body_list:
            if body.type == Body.Type.Static:
                pass
            elif body.type == Body.Type.Dynamic:
                body.pos += body.vel * dt
                body.rot += body.ang_vel * dt
                body.forces.clear()
                body.clear_torque()

            elif body.type == Body.Type.Kinematic:
                body.pos += body.vel * dt
                body.rot += body.ang_vel * dt
                body.forces.clear()
                body.clear_torque()

            elif body.type == Body.Type.Bullet:
                pass

    def solve_position_constraint(self, dt: float) -> None:
        for joint in self._joint_list:
            if joint.active:
                joint.solve_position(dt)

    @property
    def grav(self) -> Matrix:
        return self._gravity

    @grav.setter
    def grav(self, grav: Matrix) -> None:
        self._gravity = grav

    @property
    def lin_vel_damping(self) -> float:
        return self._linear_vel_damping

    @lin_vel_damping.setter
    def lin_vel_damping(self, lin_vel_damping: float) -> None:
        self._linear_vel_damping = lin_vel_damping

    @property
    def ang_vel_damping(self) -> float:
        return self._ang_vel_damping

    @ang_vel_damping.setter
    def ang_vel_damping(self, ang_vel_damping: float) -> None:
        self._ang_vel_damping = ang_vel_damping

    @property
    def lin_vel_thold(self) -> float:
        return self._linear_vel_threshold

    @lin_vel_thold.setter
    def lin_vel_thold(self, lin_vel_thold: float) -> None:
        self._linear_vel_threshold = lin_vel_thold

    @property
    def ang_vel_thold(self) -> float:
        return self._ang_vel_threshold

    @ang_vel_thold.setter
    def ang_vel_thold(self, ang_vel_thold: float) -> None:
        self._ang_vel_threshold = ang_vel_thold

    @property
    def air_fric_coeff(self) -> float:
        return self._air_fric_coeff

    @air_fric_coeff.setter
    def air_fric_coeff(self, air_fric_coeff: float) -> None:
        self._air_fric_coeff = air_fric_coeff

    @property
    def bias(self) -> float:
        return self._bias

    @bias.setter
    def bias(self, bias: float) -> None:
        self._bias = bias

    @property
    def vel_iter(self) -> int:
        return self._vel_iter

    @vel_iter.setter
    def vel_iter(self, vel_iter: int) -> None:
        self._vel_iter = vel_iter

    @property
    def pos_iter(self) -> int:
        return self._pos_iter

    @pos_iter.setter
    def pos_iter(self, pos_iter: int) -> None:
        self._pos_iter = pos_iter

    @property
    def grav_ena(self) -> bool:
        return self._grav_ena

    @grav_ena.setter
    def grav_ena(self, grav_ena: bool) -> None:
        self._grav_ena = grav_ena

    @property
    def damping_ena(self) -> bool:
        return self._damping_ena

    @damping_ena.setter
    def damping_ena(self, damping_ena: bool) -> None:
        self._damping_ena = damping_ena

    def create_body(self) -> Body:
        body: Body = Body()
        body.id = RandomGenerator.unique()
        self._body_list.append(body)
        return body

    def create_joint(
        self, prim: Union[RotationJointPrimitive, PointJointPrimitive,
                          DistanceJointPrimitive, PulleyJointPrimitive,
                          RevoluteJointPrimitive, OrientationJointPrimitive]
    ) -> Union[RotationJoint, PointJoint, DistanceJoint, PulleyJoint,
               RevoluteJoint, OrientationJoint]:

        joint: Optional[Union[RotationJoint, PointJoint, DistanceJoint,
                              PulleyJoint, RevoluteJoint,
                              OrientationJoint]] = None

        if isinstance(prim, RotationJointPrimitive):
            joint = RotationJoint(prim)
        elif isinstance(prim, PointJointPrimitive):
            joint = PointJoint(prim)
        elif isinstance(prim, DistanceJointPrimitive):
            joint = DistanceJoint(prim)
        elif isinstance(prim, PulleyJointPrimitive):
            joint = PulleyJoint(prim)
        elif isinstance(prim, RevoluteJointPrimitive):
            joint = RevoluteJoint(prim)
        elif isinstance(prim, OrientationJointPrimitive):
            joint = OrientationJoint(prim)

        joint.id = RandomGenerator.unique()
        self._joint_list.append(joint)
        return joint

    def remove_body(self, body: Body) -> None:
        for b in self._body_list:
            if body == b:
                RandomGenerator.pop(body.id)
                self._body_list.remove(body)
                break

    def remove_joint(self, joint: Joint) -> None:
        for j in self._joint_list:
            if joint == j:
                RandomGenerator.pop(joint.id)
                self._joint_list.remove(joint)

    def clear_all_bodies(self) -> None:
        self._body_list.clear()

    def clear_all_joints(self) -> None:
        self._joint_list.clear()



from __future__ import annotations
from enum import IntEnum, unique
from typing import Optional, Union

import numpy as np

from ..common.config import Config
from ..math.matrix import Matrix
from ..geometry.shape import Capsule, Ellipse, Point, Edge, Curve
from ..geometry.shape import Polygon, Sector, Shape, Circle, Rectangle


class Body():
    @unique
    class Type(IntEnum):
        Kinematic: int = 0
        Static: int = 1
        Dynamic: int = 2
        Bullet: int = 3

    class PhysicsAttribute():
        def __init__(self):
            self._pos: Matrix = Matrix([0.0, 0.0], 'vec')
            self._vel: Matrix = Matrix([0.0, 0.0], 'vec')
            self._rot: float = 0.0
            self._ang_vel: float = 0.0

        def step(self, dt: float) -> None:
            self._pos += self._vel * dt
            self._rot += self._ang_vel * dt

    def __init__(self):
        self._id: int = 0
        self._bitmask: int = 1

        self._mass: float = 0.0
        self._inv_mass: float = 0.0
        self._inertia: float = 0.0
        self._inv_inertia: float = 0.0

        self._phy_attr = Body.PhysicsAttribute()
        self._forces: Matrix = Matrix([0.0, 0.0], 'vec')
        self._torques: float = 0.0

        self._shape: Optional[Union[Point, Polygon, Rectangle, Circle, Ellipse,
                                    Edge, Curve, Capsule, Sector]] = None
        self._type = Body.Type.Static

        self._sleep: bool = False
        self._fric: float = 0.2
        self._restit: float = 0.0

    @property
    def pos(self) -> Matrix:
        return self._phy_attr._pos

    @pos.setter
    def pos(self, pos: Matrix) -> None:
        self._phy_attr._pos = pos

    @property
    def vel(self) -> Matrix:
        return self._phy_attr._vel

    @vel.setter
    def vel(self, vel: Matrix) -> None:
        self._phy_attr._vel = vel

    @property
    def rot(self) -> float:
        return self._phy_attr._rot

    @rot.setter
    def rot(self, rot: float) -> None:
        self._phy_attr._rot = rot

    @property
    def ang_vel(self) -> float:
        return self._phy_attr._ang_vel

    @ang_vel.setter
    def ang_vel(self, ang_vel: float) -> None:
        self._phy_attr._ang_vel = ang_vel

    @property
    def forces(self) -> Matrix:
        return self._forces

    @forces.setter
    def forces(self, forces: Matrix) -> None:
        self._forces = forces

    @property
    def torques(self) -> float:
        return self._torques

    @torques.setter
    def torques(self, tor: float) -> None:
        self._torques = tor

    def clear_torque(self) -> None:
        self._torques = 0.0

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, shape):
        self._shape = shape
        self.calc_inertia()

    # NOTE: need to achieve type hint
    @property
    def type(self):
        return self._type

    # NOTE: need to achieve type hint
    @type.setter
    def type(self, val):
        self._type = val

    @property
    def mass(self) -> float:
        return self._mass

    @mass.setter
    def mass(self, mass: float) -> None:
        self._mass = mass

        if np.isclose(mass, Config.Max):
            self._inv_mass = 0.0
        else:
            self._inv_mass = 0.0 if np.isclose(mass, 0) else 1.0 / mass

        self.calc_inertia()

    @property
    def inertia(self) -> float:
        return self._inertia

    # FIXME: if import AABB, will trigger loop import err
    # USE AABB.from_body static method
    # def aabb(self, factor: float = 1.0) -> AABB:
    #     prim: ShapePrimitive = ShapePrimitive()
    #     prim._xform = self._phy_attr._pos
    #     prim._rot = self._phy_attr._rot
    #     prim._shape = self._shape
    #     return AABB.from_prim(prim, factor)

    @property
    def fric(self) -> float:
        return self._fric

    @fric.setter
    def fric(self, fric: float) -> None:
        self._fric = fric

    @property
    def sleep(self) -> bool:
        return self._sleep

    @sleep.setter
    def sleep(self, sleep: bool) -> None:
        self._sleep = sleep

    @property
    def inv_mass(self) -> float:
        return self._inv_mass

    @property
    def inv_inertia(self) -> float:
        return self._inv_inertia

    @property
    def phy_attr(self) -> PhysicsAttribute:
        return self._phy_attr

    @phy_attr.setter
    def phy_attr(self, info: PhysicsAttribute):
        self._phy_attr = info

    def step_position(self, dt: float) -> None:
        self._phy_attr._pos += self._phy_attr._vel * dt
        self._phy_attr._rot += self._phy_attr._ang_vel * dt

    def apply_impulse(self, impulse: Matrix, r: Matrix) -> None:
        self._phy_attr._vel += impulse * self._inv_mass
        self._phy_attr._ang_vel += self._inv_inertia * r.cross(impulse)

    def to_local_point(self, point: Matrix) -> Matrix:
        return Matrix.rotate_mat(-self._phy_attr._rot) * (point -
                                                          self._phy_attr._pos)

    def to_world_point(self, point: Matrix) -> Matrix:
        return Matrix.rotate_mat(
            self._phy_attr._rot) * point + self._phy_attr._pos

    def to_actual_point(self, point: Matrix) -> Matrix:
        return Matrix.rotate_mat(self._phy_attr._rot) * point

    @property
    def id(self) -> int:
        return self._id

    @id.setter
    def id(self, val: int) -> None:
        self._id = val

    @property
    def bitmask(self) -> int:
        return self._bitmask

    @bitmask.setter
    def bitmask(self, bitmask: int):
        self._bitmask = bitmask

    @property
    def restit(self) -> float:
        return self._restit

    @restit.setter
    def restit(self, restit: float):
        self._restit = restit

    # NOTE: need to achieve type hint
    def calc_inertia(self):
        assert self._shape is not None
        shape_type = self._shape.type

        if shape_type == Shape.Type.Circle:
            cir: Circle = self._shape
            self._inertia = self._mass * cir.radius * cir.radius / 2.0

        elif shape_type == Shape.Type.Polygon:
            polygon: Polygon = self._shape
            center: Matrix = polygon.center()

            sum1: float = 0.0
            sum2: float = 0.0
            for i in range(len(polygon.vertices) - 1):
                n1: Matrix = polygon.vertices[i] - center
                n2: Matrix = polygon.vertices[i + 1] - center
                cross: float = np.fabs(n1.cross(n2))
                dot: float = n2.dot(n2) + n2.dot(n1) + n1.dot(n1)
                sum1 += cross * dot
                sum2 += cross

            self._inertia = self._mass * (1.0 / 6.0) * sum1 / sum2

        elif shape_type == Shape.Type.Ellipse:
            ellipse: Ellipse = self._shape
            a: float = ellipse.A()
            b: float = ellipse.B()
            self._inertia = self._mass * (a * a + b * b) * (1.0 / 5.0)

        elif shape_type == Shape.Type.Capsule:
            capsule: Capsule = self._shape
            r: float = 0.0
            h: float = 0.0
            mass_s: float = 0.0
            inertia_s: float = 0.0
            mass_c: float = 0.0
            inertia_c: float = 0.0
            volume: float = 0.0

            if capsule.width >= capsule.height:
                r = capsule.height / 2.0
                h = capsule.width - capsule.height

            else:
                r = capsule.width / 2.0
                h = capsule.height - capsule.width

            volume = np.pi * r * r + h * 2 * r
            rho: float = self._mass / volume
            mass_s = rho * np.pi * r * r
            mass_c = rho * h * 2.0 * r
            inertia_c = (1.0 / 12.0) * mass_c * (h * h + (2.0 * r) * (2.0 * r))
            inertia_s = mass_s * r * r * 0.5
            self._inertia = inertia_c + inertia_s + mass_s * (
                3.0 * r + 2.0 * h) * h / 8.0

        elif shape_type == Shape.Type.Sector:
            sector: Sector = self._shape
            self._inertia = self._mass * (
                sector.span - np.sin(sector.span)
            ) * sector.radius * sector.radius / 4.0 * sector.span

        if np.isclose(self._mass, Config.Max):
            self._inv_inertia = 0.0
        else:
            self._inv_inertia = 1.0 / self._inertia if not np.isclose(
                self._inertia, 0) else 0.0



import os
import sys
import shutil
import runpy
import argparse
import subprocess
from pathlib import Path
from typing import Dict

import colorama as cra


def registerableCLI(cls):
    """Class decorator to register methodss with @register into a set."""
    cls._registered_commands = set([])
    for name in dir(cls):
        method = getattr(cls, name)
        if hasattr(method, '_registered'):
            cls._registered_commands.add(name)
    return cls


def register(func):
    """Method decorator to register CLI commands."""
    func._registered = True
    return func


@registerableCLI
class TaichiGAMEMain():
    def __init__(self, test_mode: bool = False):
        cra.init()
        self._banner = cra.Fore.GREEN + f"\n{'#' * 70}\n"
        self._banner += '##  TaichiGAME: GPU Accelerated Motion Engine '
        self._banner += 'based on Taichi Lang  ##'
        self._banner += f"\n{'#' * 70}"
        print(self._banner)

        parser = argparse.ArgumentParser(description="TaichiGAME CLI",
                                         usage=self._usage())
        parser.add_argument('command',
                            help="command from the above list to run")

        # Flag for unit testing
        self._test_mode = test_mode
        self._main_parser = parser

    def _usage(self) -> str:
        """Compose deterministic usage message based on registered_commands."""
        # TODO: add some color to commands
        msg = "\n"
        space = 20
        for command in sorted(self._registered_commands):  # type: ignore # pylint: disable=E1101
            msg += f"{command}{' ' * (space - len(command))}|-> {getattr(self, command).__doc__}\n"
        return msg

    @staticmethod
    def _exec_python_file(filename: str):
        """Execute a Python file based on filename."""
        # TODO: do we really need this?
        subprocess.call([sys.executable, filename] + sys.argv[1:])

    @staticmethod
    def _get_examples_dir() -> Path:
        """Get the path to the examples directory."""
        root_dir = os.path.join(
            os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

        examples_dir = Path(root_dir) / 'TaichiGAME/examples'
        return examples_dir

    @staticmethod
    def _get_available_examples() -> Dict[str, Path]:
        """Get a set of all available example names."""
        examples_dir = TaichiGAMEMain._get_examples_dir()
        all_examples = examples_dir.rglob('*.py')
        all_example_names = {f.stem: f.parent for f in all_examples}

        res: Dict[str, Path] = {}
        for v in all_example_names:
            if v == 'main' or v == '__init__':
                continue

            res[v] = all_example_names[v]

        return res

    @staticmethod
    def _example_choices_type(choices):
        def support_choice_with_dot_py(choice):
            if choice.endswith('.py') and choice.split('.')[0] in choices:
                # try to find and remove python file extension
                return choice.split('.')[0]
            return choice

        return support_choice_with_dot_py

    def __call__(self):
        # Print help if no command provided
        if len(sys.argv[1:2]) == 0:
            self._main_parser.print_help()
            return 1

        # Parse the command
        args = self._main_parser.parse_args(sys.argv[1:2])

        if args.command not in self._registered_commands:  # type: ignore # pylint: disable=E1101
            # TODO: do we really need this?
            if args.command.endswith(".py"):
                TaichiGAMEMain._exec_python_file(args.command)
            else:
                print(f"{args.command} is not a valid command!")
                self._main_parser.print_help()
            return 1

        return getattr(self, args.command)(sys.argv[2:])

    @register
    def example(self, arguments: list = sys.argv[2:]):
        """Run an example by name (or name.py)"""
        choices = TaichiGAMEMain._get_available_examples()

        parser = argparse.ArgumentParser(prog='python3 -m TaichiGAME example',
                                         description=f"{self.example.__doc__}")

        parser.add_argument(
            "name",
            help="Name of an example (supports .py extension too)\n",
            type=TaichiGAMEMain._example_choices_type(choices.keys()),
            choices=sorted(choices.keys()))
        parser.add_argument(
            '-p',
            '--print',
            required=False,
            dest='print',
            action='store_true',
            help="Print example source code instead of running it")
        parser.add_argument(
            '-P',
            '--pretty-print',
            required=False,
            dest='pretty_print',
            action='store_true',
            help="Like --print, but print in a rich format with line numbers")
        parser.add_argument(
            '-s',
            '--save',
            required=False,
            dest='save',
            action='store_true',
            help="Save source code to current directory instead of running it")

        # TODO: Pass the arguments to downstream correctly(#3216).
        args = parser.parse_args(arguments)
        examples_dir = TaichiGAMEMain._get_examples_dir()
        target = str(
            (examples_dir / choices[args.name] / f"{args.name}.py").resolve())
        # path for examples needs to be modified for implicit relative imports
        sys.path.append(str((examples_dir / choices[args.name]).resolve()))

        # Short circuit for testing
        if self._test_mode:
            return args

        if args.save:
            print(f"Saving example {args.name} to current directory...")
            shutil.copy(target, '.')
            return 0

        if args.pretty_print:
            return 0

        if args.print:
            with open(target, encoding='utf-8') as f:
                print(f.read())
            return 0

        print(f"Running example {args.name} ...")

        runpy.run_path(target, run_name='__main__')

        return None


def main():
    cli = TaichiGAMEMain()

    return cli()


if __name__ == '__main__':
    sys.exit(main())



from __future__ import annotations
from typing import List, Tuple, Union

import numpy as np


# now just for the 2x2 mat or 1x2 vec
class Matrix():
    def __init__(self,
                 arr: Union[List[float], np.ndarray],
                 data_type: str = 'mat',
                 row: int = 2,
                 col: int = 2):

        self._data_type: str = data_type
        self._val: np.ndarray = np.array(arr).reshape(
            row, 1 if data_type == 'vec' else col)

    # unary operator
    def __neg__(self) -> Matrix:
        return Matrix(-self._val, self._data_type)

    def __pos__(self) -> Matrix:
        return Matrix(self._val, self._data_type)

    def __invert__(self):
        raise NotImplementedError

    # binary operator
    def __add__(self, other: Union[float, int, Matrix]) -> Matrix:
        if isinstance(other, float) or isinstance(other, int):
            return Matrix(self._val + other, self._data_type)
        else:
            return Matrix(self._val + other._val, self._data_type)

    def __sub__(self, other: Union[float, int, Matrix]) -> Matrix:
        if isinstance(other, float) or isinstance(other, int):
            return Matrix(self._val - other, self._data_type)
        else:
            return Matrix(self._val - other._val, self._data_type)

    def __mul__(self, other: Union[float, int, Matrix]) -> Matrix:
        if isinstance(other, float) or isinstance(other, int):
            return Matrix(self._val * other, self._data_type)
        else:
            assert self._val.shape[1] == other._val.shape[0]
            return Matrix(self._val @ other._val, other._data_type)

    def __truediv__(self, other: float) -> Matrix:
        assert not np.isclose(other, 0)
        return Matrix(self._val / other, self._data_type)

    def __floordiv__(self, other):
        raise NotImplementedError

    def __mod__(self, other):
        raise NotImplementedError

    def __pow__(self, other):
        raise NotImplementedError

    def __rshift__(self, other):
        raise NotImplementedError

    def __lshift__(self, other):
        raise NotImplementedError

    def __and__(self, other):
        raise NotImplementedError

    def __or__(self, other):
        raise NotImplementedError

    def __xor__(self, other):
        raise NotImplementedError

    # comparsion operator
    def __lt__(self, other):
        raise NotImplementedError

    def __gt__(self, other):
        raise NotImplementedError

    def __le__(self, other):
        raise NotImplementedError

    def __ge__(self, other):
        raise NotImplementedError

    def __eq__(self, other) -> bool:
        return np.isclose(self._val, other._val).all()

    def __ne__(self, other) -> bool:
        return not np.isclose(self._val, other._val).all()

    # assignment operator
    def __isub__(self, other: Union[float, int, Matrix]) -> Matrix:
        if isinstance(other, float) or isinstance(other, int):
            self._val -= other
        else:
            self._val -= other._val
        return self

    def __iadd__(self, other: Union[float, int, Matrix]) -> Matrix:
        if isinstance(other, float) or isinstance(other, int):
            self._val += other
        else:
            self._val += other._val
        return self

    def __imul__(self, other: Union[float, int, Matrix]) -> Matrix:
        if isinstance(other, float) or isinstance(other, int):
            self._val *= other
        else:
            assert self._val.shape[1] == other._val.shape[0]
            self._val = self._val @ other._val

        return self

    def __idiv__(self, other: float) -> Matrix:
        assert not np.isclose(other, 0)
        self._val /= other
        return self

    def __ifloordiv__(self, other):
        raise NotImplementedError

    def __imod__(self, other):
        raise NotImplementedError

    def __ipow__(self, other):
        raise NotImplementedError

    def __irshift__(self, other):
        raise NotImplementedError

    def __ilshift__(self, other):
        raise NotImplementedError

    def __iand__(self, other):
        raise NotImplementedError

    def __ior__(self, other):
        raise NotImplementedError

    def __ixor__(self, other):
        raise NotImplementedError

    def __str__(self) -> str:
        res: str = ''
        for i in self._val:
            res += str(i) + '\n'

        return res

    @property
    def x(self) -> float:
        '''extern interface for the 2d vector's x pos

        Returns
        -------
        float
            x pos of the vector
        '''
        return self._val[0, 0]

    @x.setter
    def x(self, val: float):
        self._val[0, 0] = val

    @property
    def y(self) -> float:
        '''extern interface for the 2d vector's y pos

        Returns
        -------
        float
            y pos of the vector
        '''
        if self._val.shape == (2, 1):
            return self._val[1, 0]
        elif self._val.shape == (1, 2):
            return self._val[0, 1]
        else:
            raise ValueError

    @y.setter
    def y(self, val: float):
        self._val[1, 0] = val

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._val.shape

    @property
    def size(self) -> int:
        return self._val.size

    @property
    def row1(self) -> Matrix:
        assert self._val.shape == (2, 2)
        return Matrix(self._val[0], 'vec')

    @property
    def row2(self) -> Matrix:
        assert self._val.shape == (2, 2)
        return Matrix(self._val[1], 'vec')

    def reshape(self, row: int, col: int) -> Matrix:
        self._val = self._val.reshape(row, col)
        return self

    def value(self, row: int = 0, col: int = 0) -> float:
        assert self._val.shape == (2, 2)
        assert 0 <= row <= self._val.shape[0]
        assert 0 <= col <= self._val.shape[1]
        return self._val[row, col]

    def determinant(self) -> float:
        assert self._val.shape == (2, 2)
        return np.linalg.det(self._val)

    def transpose(self) -> Matrix:
        self._val = self._val.T
        return self

    def invert(self) -> Matrix:
        assert self._val.shape == (2, 2)
        self._val = np.linalg.inv(self._val)
        return self

    def skew_symmetric_mat(self, vec: Matrix) -> Matrix:
        assert self._val.shape == (2, 2)
        return Matrix([0, -vec._val[1, 0], vec._val[0, 0], 0])

    def identity_mat(self) -> Matrix:
        assert self._val.shape == (2, 2)
        return Matrix([1, 0, 0, 1])

    def len_square(self) -> float:
        return np.square(self._val).sum()

    def len(self) -> float:
        return np.sqrt(self.len_square())

    def theta(self) -> float:
        assert self._val.shape == (2, 1)
        assert not np.isclose(self._val[0, 0], 0)
        return np.arctan2(self._val[1, 0], self._val[0, 0])

    def set_value(self, val: Union[List[float], Matrix]) -> Matrix:
        if isinstance(val, list):
            self._val = np.array(val).reshape(self._val.shape)
        elif isinstance(val, Matrix):
            self._val = val._val
        return self

    def clear(self) -> Matrix:
        if self._val.shape == (2, 2):
            self.set_value([0.0, 0.0, 0.0, 0.0])
        else:
            self.set_value([0.0, 0.0])

        return self

    def negate(self) -> Matrix:
        self._val = -self._val
        return self

    def negative(self) -> Matrix:
        return Matrix(-self._val, self._data_type)

    def swap(self, other: Matrix) -> Matrix:
        assert self._data_type == other._data_type
        assert self._val.shape == other._val.shape
        self._val, other._val = other._val, self._val

        return self

    def normalize(self) -> Matrix:
        assert not np.isclose(self.len(), 0)
        self._val /= self.len()
        return self

    def normal(self) -> Matrix:
        assert not np.isclose(self.len(), 0)
        return Matrix(self._val / self.len(), self._data_type)

    def is_origin(self) -> bool:
        assert self._val.shape == (2, 1)
        return np.isclose(self._val, [0, 0]).all()

    def dot(self, other: Matrix) -> float:
        assert self._val.shape == (2, 1)
        assert other._val.shape == (2, 1)
        return np.dot(self._val.T, other._val)[0, 0]

    def cross(self, other: Matrix) -> float:
        assert self._val.shape == (2, 1)
        assert other._val.shape == (2, 1)
        # NOTE: same as the cross_product method
        return np.cross(self._val.reshape(2), other._val.reshape(2)).tolist()

    def perpendicular(self) -> Matrix:
        assert self._val.shape == (2, 1)
        return Matrix([-self._val[1, 0], self._val[0, 0]], self._data_type)

    @staticmethod
    def dot_product(veca: Matrix, vecb: Matrix) -> float:
        assert veca._val.shape == (2, 1)
        assert vecb._val.shape == (2, 1)
        return np.dot(veca._val.T, vecb._val)[0, 0]

    @staticmethod
    def cross_product(veca: Matrix, vecb: Matrix) -> float:
        assert veca._val.shape == (2, 1)
        assert vecb._val.shape == (2, 1)
        # NOTE: just hack this impl to output scalar val otherwise vector
        # to pass mypy check
        return np.cross(veca._val.reshape(2), vecb._val.reshape(2)).tolist()

    @staticmethod
    def cross_product2(lhs: Union[Matrix, float], rhs: Union[Matrix,
                                                             float]) -> Matrix:
        if isinstance(lhs, float) and isinstance(rhs, Matrix):
            assert rhs._val.shape == (2, 1)
            return Matrix([-rhs.y, rhs.x], 'vec') * lhs

        elif isinstance(lhs, Matrix) and isinstance(rhs, float):
            assert lhs._val.shape == (2, 1)
            return Matrix([lhs.y, -lhs.x], 'vec') * rhs
        else:
            raise TypeError

    @staticmethod
    def rotate_mat(radian: float) -> Matrix:
        res: List[float] = []
        cos_val: float = np.cos(radian)
        sin_val: float = np.sin(radian)
        res.append(cos_val)
        res.append(-sin_val)
        res.append(sin_val)
        res.append(cos_val)

        return Matrix(res)
